//
// Filter.cpp
//   Implementation for HAIL image filter classes
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include "HAIL.h"
#include "HAIL_private.h"
#include "Buffer.h"
#include "ATIExtensions.h"

using namespace HAIL;

#define for if(0); else for

//
//
//////////////////////////////////////////////////////////////////////
ImageFilter::~ImageFilter()
{
}

//
//
//////////////////////////////////////////////////////////////////////
Convolve2D::Convolve2D( int width, int height, float *filter) : ImageFilter( width, height)
{
   m_data = new float[width*height];

   m_shaders = NULL;
   m_passes = 0;
   m_dirty = true;

   if (filter)
   {
      memcpy( m_data, filter, sizeof(float)*width*height);
   }
   else
   {
      memset( m_data, 0, sizeof(float)*width*height);
   }
}

//
//
//////////////////////////////////////////////////////////////////////
Convolve2D::~Convolve2D()
{
   delete []m_data;
   if (m_shaders)
      glDeleteProgramsARB( m_passes, m_shaders);
   delete []m_shaders;
}

//
//
//////////////////////////////////////////////////////////////////////
void Convolve2D::SetComponent( int x, int y, float val)
{
   m_data[x*GetWidth()+y] = val;
   m_dirty = true;
}

//
//
//////////////////////////////////////////////////////////////////////
float Convolve2D::GetComponent( int x, int y)
{
   return m_data[x*GetWidth()+y];
}

//
//
//////////////////////////////////////////////////////////////////////
bool Convolve2D::Seperable()
{
   return false;
}

//
// Compile a convolution into a set of fragment shader passes
//////////////////////////////////////////////////////////////////////
void Convolve2D::Compile()
{
   if (!m_dirty)
   {
      return;
   }

   if (m_shaders)
   {
      
      glDeleteProgramsARB( m_passes, m_shaders);
      delete []m_shaders;
      m_shaders = 0;
   }

   m_passes = (GetWidth()*GetHeight())/9;
   m_passes += ((GetWidth()*GetHeight())%9 > 0);

   m_shaders = new unsigned int[m_passes];
   glGenProgramsARB( m_passes, m_shaders);

   float x = -GetWidth()/2, y = -GetHeight()/2;

   //create the shaders
   for (int ii=0; ii<m_passes; ii++)
   {
      strstream shader;
      

      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, m_shaders[ii]);

      //this creates a worst case shader each time
      CreateShaderPreamble( shader, 1, 10);
      shader << "PARAM c0 = program.local[0];\n";

      //replace this with position dependent values
      for (int gg=0; (gg<9) && ((GetWidth()*GetHeight())>(ii*9+gg)); gg++)
      {
         shader << "PARAM c" << gg+1 <<" = { " << x << ", " << y << ", " << m_data[ii*9+gg] << "};\n";
         if ( (ii*9+gg)%GetWidth() == GetWidth()-1)
         {
            x = -GetWidth()/2;
            y += 1.0f;
         }
         else
         {
            x += 1.0f;
         }
      }

      shader << "\n";

      //compute filter locations
      for (int gg=0; (gg<9) && ((GetWidth()*GetHeight())>(ii*9+gg)); gg++)
      {
         shader << "MAD r" << gg << ".xy, c0, c" << gg+1 << ", t0;\n"; 
      }

      shader << "\n";

      //fetch
      for (int gg=0; (gg<9) && ((GetWidth()*GetHeight())>(ii*9+gg)); gg++)
      {
         shader << "TEX r" << gg << ", r" << gg <<", texture[0], 2D;\n";
      }

      shader << "\n";

      if (ii>0)
      {
         shader << "TEX r9, t0, texture[1], 2D;\n";
      }

      shader << "\n";

      //sum
      if (ii>0)
      {
         shader << "MAD r0, r0, c1.b, r9;\n";
      }
      else
      {
         shader << "MUL r0, r0, c1.b;\n";
      }
      for (int gg=1; (gg<9) && ((GetWidth()*GetHeight())>(ii*9+gg)); gg++)
      {
         shader << "MAD r0, r" << gg << ", c" << gg+1 <<".b, r0;\n";
      }

      shader << "MOV res, r0;\n\n";

      shader << "END\n\n" << '\0';

      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());
      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
         printf("Error in shader: %s\n\n", dbg);
         printf("\n\n#\n# Shader %d\n#\n\n%s", ii, shader.str());
      }

   }

   m_dirty = false;
}

//
//
//////////////////////////////////////////////////////////////////////
int Convolve2D::Passes()
{

   return m_passes;
}

//
//
//////////////////////////////////////////////////////////////////////
bool Convolve2D::BindPass( int n)
{
   if (m_dirty)
      return false;

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, m_shaders[n]);

   return true;
}

//
//
//////////////////////////////////////////////////////////////////////
Convolve2DSeperable::Convolve2DSeperable( int width, int height, float *hor, float *vert) : ImageFilter( width, height)
{
   m_hData = new float[width];
   m_vData = new float[height];

   if (hor)
   {
      memcpy( m_hData, hor, sizeof(float)*width);
   }
   else
   {
      memset( m_hData, 0, sizeof(float)*width);
   }

   if (vert)
   {
      memcpy( m_vData, vert, sizeof(float)*height);
   }
   else
   {
      memset( m_vData, 0, sizeof(float)*height);
   }

   m_hShaders = 0;
   m_vShaders = 0;
   m_hPasses = 0;
   m_vPasses = 0;
   m_dirty = true;
}

//
//
//////////////////////////////////////////////////////////////////////
Convolve2DSeperable::~Convolve2DSeperable()
{
   delete []m_hData;
   delete []m_vData;
   if (m_hShaders)
      glDeleteProgramsARB( m_hPasses, m_hShaders);
   delete []m_hShaders;
   if (m_vShaders)
      glDeleteProgramsARB( m_vPasses, m_vShaders);
   delete []m_vShaders;
}

//
//
//////////////////////////////////////////////////////////////////////
void Convolve2DSeperable::SetHorComponent( int x, float val)
{
   m_hData[x] = val;
   m_dirty = true;
}

//
//
//////////////////////////////////////////////////////////////////////
void Convolve2DSeperable::SetVertComponent( int y, float val)
{
   m_vData[y] = val;
   m_dirty = true;
}

//
//
//////////////////////////////////////////////////////////////////////
float Convolve2DSeperable::GetHorComponent( int x)
{
   return m_hData[x];
}

//
//
//////////////////////////////////////////////////////////////////////
float Convolve2DSeperable::GetVertComponent( int y)
{
   return m_vData[y];
}

//
//
//////////////////////////////////////////////////////////////////////
bool Convolve2DSeperable::Seperable()
{
   return true;
}

//
//
//////////////////////////////////////////////////////////////////////
void Convolve2DSeperable::Compile()
{
   if (!m_dirty)
   {
      return;
   }

   if (m_hShaders)
   {
      
      glDeleteProgramsARB( m_hPasses, m_hShaders);
      delete []m_hShaders;
      m_hShaders = 0;
   }

   if (m_vShaders)
   {
      
      glDeleteProgramsARB( m_vPasses, m_vShaders);
      delete []m_vShaders;
      m_vShaders = 0;
   }

   m_hPasses = GetWidth()/9;
   m_hPasses += ((GetWidth()%9) > 0);
   m_vPasses = GetHeight()/9;
   m_vPasses += ((GetHeight()%9) > 0);

   m_hShaders = new unsigned int[m_hPasses];
   m_vShaders = new unsigned int[m_vPasses];
   glGenProgramsARB( m_hPasses, m_hShaders);
   glGenProgramsARB( m_vPasses, m_vShaders);

   float x = -GetWidth()/2, y = -GetHeight()/2;

   //create the shaders for the horizontal passes
   for (int ii=0; ii<m_hPasses; ii++)
   {
      strstream shader;
      

      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, m_hShaders[ii]);

      //this creates a worst case shader each time
      CreateShaderPreamble( shader, 1, 10);
      shader << "PARAM c0 = program.local[0];\n";

      //replace this with position dependent values
      for (int gg=0; (gg<9) && (GetWidth()>(ii*9+gg)); gg++)
      {
         shader << "PARAM c" << gg+1 <<" = { " << x << ", 0.0, " << m_hData[ii*9+gg] << "};\n";
         
         x += 1.0f;
      }

      shader << "\n";

      //compute filter locations
      for (int gg=0; (gg<9) && (GetWidth()>(ii*9+gg)); gg++)
      {
         shader << "MAD r" << gg << ".xy, c0, c" << gg+1 << ", t0;\n"; 
      }

      shader << "\n";

      //fetch
      for (int gg=0; (gg<9) && (GetWidth()>(ii*9+gg)); gg++)
      {
         shader << "TEX r" << gg << ", r" << gg <<", texture[0], 2D;\n";
      }

      shader << "\n";

      if (ii>0)
      {
         shader << "TEX r9, t0, texture[1], 2D;\n";
      }

      shader << "\n";

      //sum
      if (ii>0)
      {
         shader << "MAD r0, r0, c1.b, r9;\n";
      }
      else
      {
         shader << "MUL r0, r0, c1.b;\n";
      }
      for (int gg=1; (gg<9) && (GetWidth()>(ii*9+gg)); gg++)
      {
         shader << "MAD r0, r" << gg << ", c" << gg+1 <<".b, r0;\n";
      }

      shader << "MOV res, r0;\n\n";

      shader << "END\n\n" << '\0';

      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());
      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
         printf("Error in shader: %s\n\n", dbg);
         printf("\n\n#\n# Horizontal Shader %d\n#\n\n%s", ii, shader.str());
      }

   }

   //create the shaders for the vertical passes
   for (int ii=0; ii<m_vPasses; ii++)
   {
      strstream shader;
      

      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, m_vShaders[ii]);

      //this creates a worst case shader each time
      CreateShaderPreamble( shader, 1, 10);
      shader << "PARAM c0 = program.local[0];\n";

      //replace this with position dependent values
      for (int gg=0; (gg<9) && (GetHeight()>(ii*9+gg)); gg++)
      {
         shader << "PARAM c" << gg+1 <<" = { 0.0, " << y << ", " << m_vData[ii*9+gg] << "};\n";
         
         y += 1.0f;
         
      }

      shader << "\n";

      //compute filter locations
      for (int gg=0; (gg<9) && (GetHeight()>(ii*9+gg)); gg++)
      {
         shader << "MAD r" << gg << ".xy, c0, c" << gg+1 << ", t0;\n"; 
      }

      shader << "\n";

      //fetch
      for (int gg=0; (gg<9) && (GetHeight()>(ii*9+gg)); gg++)
      {
         shader << "TEX r" << gg << ", r" << gg <<", texture[0], 2D;\n";
      }

      shader << "\n";

      if (ii>0)
      {
         shader << "TEX r9, t0, texture[1], 2D;\n";
      }

      shader << "\n";

      //sum
      if (ii>0)
      {
         shader << "MAD r0, r0, c1.b, r9;\n";
      }
      else
      {
         shader << "MUL r0, r0, c1.b;\n";
      }
      for (int gg=1; (gg<9) && (GetHeight()>(ii*9+gg)); gg++)
      {
         shader << "MAD r0, r" << gg << ", c" << gg+1 <<".b, r0;\n";
      }

      shader << "MOV res, r0;\n\n";

      shader << "END\n\n" << '\0';

      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());
      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
         printf("Error in shader: %s\n\n", dbg);
         printf("\n\n#\n# Vertical Shader %d\n#\n\n%s", ii, shader.str());
      }
   }


   m_dirty = false;
}

//
//
//////////////////////////////////////////////////////////////////////
int Convolve2DSeperable::Passes()
{

   return m_hPasses + m_vPasses;
}

//
//
//////////////////////////////////////////////////////////////////////
int Convolve2DSeperable::VerticalPasses()
{
   return m_vPasses;
}

//
//
//////////////////////////////////////////////////////////////////////
int Convolve2DSeperable::HorizontalPasses()
{
   return m_hPasses;
}

//
//
//////////////////////////////////////////////////////////////////////
bool Convolve2DSeperable::BindHorizontalPass( int n)
{
   if (!m_dirty)
   {
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, m_hShaders[n]);
      return true;
   }
   return false;
}

//
//
//////////////////////////////////////////////////////////////////////
bool Convolve2DSeperable::BindVerticalPass( int n)
{
   if (!m_dirty)
   {
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, m_vShaders[n]);
      return true;
   }
   return false;
}