//
// HAIL - Hardware Accelerated Image Library
//   - based on OpenGL
//   - uses pBuffers and fragment programs to perform image transforms
//   - primarily designed as a prototype to show the possibilities
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef HAIL_H
#define HAIL_H

#define HAIL_VERSION  001

namespace HAIL
{
   //
   // Enums for mode setting
   /////////////////////////////////////////////////////////////////////////////
   enum WrapMode
   {
      CLAMP,
      REPEAT,
      BORDER
   };

   enum SampleMode
   {
      NEAREST,
      LINEAR,
      CUBIC
   };

   enum ImageType
   {
      MONOCHROME,
      LUMINANCE_ALPHA,
      RGB,
      RGBA
   };

   enum ApplyMode
   {
      MODULATE,
      ADD,
      SUBTRACT,
      REPLACE,
      NOOP
   };

   //
   // General Support Functions
   /////////////////////////////////////////////////////////////////////////////
   
   //
   // Initialize the support code returns true on success
   ///////////////////////////////////////////////////////////////////
   bool Init();

   //
   // Shutdown the library and free shared resources
   ///////////////////////////////////////////////////////////////////
   void Shutdown();

   //
   // Classes implementing the core of the library
   /////////////////////////////////////////////////////////////////////////////

   //
   // Filter that can be applied to an Image
   ///////////////////////////////////////////////////////////////////
   class ImageFilter
   {
      int m_width, m_height;

      public:
         ImageFilter( int width, int height) : m_width(width), m_height(height) {};
         virtual ~ImageFilter();

         virtual bool Seperable() = 0;
         virtual void Compile() = 0;
         virtual int Passes() = 0;

         int GetWidth() {return m_width;};
         int GetHeight() {return m_height;};

   };

   //
   // Generic 2D Convolution filter 
   ///////////////////////////////////////////////////////////////////
   class Convolve2D : public ImageFilter
   {
      float *m_data;

      unsigned int *m_shaders;
      int m_passes;
      bool m_dirty;

      public:
         Convolve2D( int width, int height, float *filter = 0);
         ~Convolve2D();

         void SetComponent( int x, int y, float val);
         float GetComponent( int x, int y);

         bool Seperable();
         void Compile();
         int Passes();
         bool BindPass( int n);
   };

   //
   // Seperable 2D convolution  
   ///////////////////////////////////////////////////////////////////
   class Convolve2DSeperable : ImageFilter
   {
      float *m_hData;
      float *m_vData;

      unsigned int *m_hShaders;
      unsigned int *m_vShaders;
      int m_hPasses;
      int m_vPasses;
      bool m_dirty;

      public:
         Convolve2DSeperable( int width, int height, float *hor = 0, float *vert = 0);
         ~Convolve2DSeperable();

         void SetHorComponent( int x, float val);
         void SetVertComponent( int y, float val);
         float GetHorComponent( int x);
         float GetVertComponent( int y);

         bool Seperable();
         void Compile();
         int Passes();
         int VerticalPasses();
         int HorizontalPasses();
         bool BindHorizontalPass( int n);
         bool BindVerticalPass( int n);
   };

   //
   //
   ///////////////////////////////////////////////////////////////////
   class Image2D
   {
      public:
         Image2D();
         ~Image2D();

         void SetBorderMode( WrapMode mode);
         WrapMode GetBorderMode();

         void Crop( int x, int y, int width, int height);
         void Resize( SampleMode mode, int x, int y);

         void ApplyFilter( ImageFilter *filter);
         void ApplyImage( ApplyMode mode, Image2D *image);


         void FFT();

         void IFFT();

         void Copy( Image2D *dest);

         int GetWidth();
         int GetHeight();

         bool ImageLoad( int width, int height, ImageType type, unsigned char* data);
         bool LoadImageFromTexture( int width, int height, ImageType type, unsigned int tex);

         ImageType GetType();

         unsigned int GetTexHandle() {return m_fTexHandle;};

      private:

         //shader stuff

         //helper functions
         void SetParameters();
         void ApplySeperableFilter( ImageFilter *filter);
         void ApplyInseperableFilter( ImageFilter *filter);

         //
         //static data for supporting common shaders
         //
         static void InitFFTShaders();
         static void InitInterpolationShaders();
         static void InitApplyShaders();

         //FFT shaders
         static unsigned int hScramble;
         static unsigned int vScramble;
         static unsigned int hButterfly;
         static unsigned int vButterfly;
         static unsigned int rescale;
         
         //resample shaders
         static unsigned int nearest;
         static unsigned int linear;
         static unsigned int cubic;

         //apply shaders
         static unsigned int add;
         static unsigned int subtract;
         static unsigned int modulate;
         static unsigned int replace;

         //make these illegal
         operator=( const Image2D &rhs) {};
         Image2D( const Image2D& src) {};


         //image info
         int m_width;
         int m_height;
         int m_pow2Width;
         int m_pow2Height;
         ImageType m_type;
         WrapMode m_wrap;

         //hardware info
         unsigned int m_fTexHandle;
   };
};


#endif //HAIL_H