//
// Image2D.cpp
//  Implementation for all of the 2D Image operations except FFT
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "HAIL.h"
#include "HAIL_private.h"

#include "Buffer.h"

#include "ATIExtensions.h"

using namespace HAIL;

#define for if(0);else for

//
//
//////////////////////////////////////////////////////////////////////////////////////////
static int Pow2( int val)
{
   int ret = 1;
   
   while ( ret < val)
   {
      ret <<= 1;
   }

   return ret;
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
static void CheckErr( int lineno)
{
   GLuint err = glGetError();

   switch(err)
   {
      case GL_INVALID_ENUM:
         printf( "Invalid Enum: %d\n", lineno);
         break;
      case GL_INVALID_VALUE:
         printf( "Invalid Value: %d\n", lineno);
         break;
      case GL_INVALID_OPERATION:
         printf( "Invalid Operation: %d\n", lineno);
         break;
      case GL_NO_ERROR:
         break;
      default:
         printf( "Generic Error: %d\n", lineno);
   };
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
Image2D::Image2D()
{

   //image info
   m_width = 0;
   m_height = 0;
   m_pow2Width = 0;
   m_pow2Height = 0;
   m_type = MONOCHROME;
   m_wrap = REPEAT;

   //hardware info
   m_fTexHandle = 0;

   InitFFTShaders();
   InitInterpolationShaders();
   InitApplyShaders();
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
Image2D::~Image2D()
{
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::SetBorderMode( WrapMode mode)
{
   m_wrap = mode;

   if (m_fTexHandle)
   {
      BindBuffer( FLOAT_BUFFER);

      glBindTexture( GL_TEXTURE_2D, m_fTexHandle);

      switch (m_wrap)
      {
         case REPEAT:
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
            break;
         case CLAMP:
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
            break;
         case BORDER:
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER_ARB);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER_ARB);
            break;
      };

      UnbindBuffer(FLOAT_BUFFER);
   }
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
WrapMode Image2D::GetBorderMode()
{
   return m_wrap;
}


//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::Crop( int x, int y, int width, int height)
{
   BindBuffer( FLOAT_BUFFER);

   //set viewport and proj matrix
   glViewport( 0, 0, width, height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0f, 1.0f, 0.0f, 1.0f);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   //turn on textures and bind
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable(GL_TEXTURE_2D);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glMatrixMode( GL_TEXTURE);
   glLoadIdentity();
   glScalef( 1.0f/(float)m_pow2Width, 1.0f/(float)m_pow2Height, 1.0f);

   //draw the primitive
   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, x, y);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, x+width, y);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, x+width, y+height);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, x, y+height);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   m_pow2Width = Pow2(width);
   m_pow2Height = Pow2(height);
   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, m_pow2Width, m_pow2Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, width, height);

   m_width = width;
   m_height = height;

   UnbindBuffer( FLOAT_BUFFER);
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::Resize( SampleMode mode, int x, int y)
{
   BindBuffer( FLOAT_BUFFER);

   CheckErr(__LINE__);

   //set viewport and proj matrix
   glViewport( 0, 0, x, y);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0f, 1.0f, 0.0f, 1.0f);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   //turn on textures and bind
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable(GL_TEXTURE_2D);

   //create/configure fragment program
   strstream shader;
   
   switch (mode)
   {
      case NEAREST:
         
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::nearest);
         break;
      case LINEAR:
         //bind filter masks
         glActiveTextureARB( GL_TEXTURE1_ARB);
         BindUtilityTexture( LINEAR_TEXTURE);
         glEnable( GL_TEXTURE_2D);
         
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::linear);
         break;
      case CUBIC:
         //bind filter masks
         glActiveTextureARB( GL_TEXTURE1_ARB);
         BindUtilityTexture( CUBIC_TEXTURE);
         glEnable( GL_TEXTURE_2D);
         
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::cubic);
         break;
   };
   
   CheckErr(__LINE__);

   glProgramLocalParameter4fARB( GL_FRAGMENT_PROGRAM_ARB, 0, 1.0f/m_pow2Width, 1.0f/m_pow2Height, 0.0f, 0.0f);

   
   glEnable(GL_FRAGMENT_PROGRAM_ARB);


   glBegin(GL_QUADS);
   
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE2_ARB, 0.0f, 0.0f);
         glVertex3f( 0.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE2_ARB, (float)m_width/(float)m_pow2Width, 0.0f);
         glVertex3f( 1.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE2_ARB, (float)m_width/(float)m_pow2Width, (float)m_height/(float)m_pow2Height);
         glVertex3f( 1.0f, 1.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE2_ARB, 0.0f, (float)m_height/(float)m_pow2Height);
         glVertex3f( 0.0f, 1.0f, -0.5f);

   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, 0);

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glDisable( GL_TEXTURE_2D);
   glBindTexture( GL_TEXTURE_2D, 0);

   glActiveTextureARB( GL_TEXTURE0_ARB);

   //update size
   m_width = x;
   m_height = y;
   m_pow2Width = Pow2(m_width);
   m_pow2Height = Pow2(m_height);

   //float next
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_width, m_height);

   glBindTexture(GL_TEXTURE_2D, 0);

   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE2_ARB);
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glMatrixMode(GL_MODELVIEW);

   UnbindBuffer(FLOAT_BUFFER);
}


//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::ApplyFilter( ImageFilter *filter)
{
   BindBuffer( FLOAT_BUFFER);

   //generic setup
   glViewport( 0, 0, m_width, m_height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   //main image setup
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable( GL_TEXTURE_2D);
   glMatrixMode( GL_TEXTURE);
   glLoadIdentity();
   glScalef( (float)m_width/(float)m_pow2Width, (float)m_height/(float)m_pow2Height, 1.0f);

   glMatrixMode(GL_MODELVIEW);

   if (filter->Seperable())
   {
      ApplySeperableFilter(filter);
   }
   else
   {
      ApplyInseperableFilter(filter);
   }

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glMatrixMode( GL_TEXTURE);
   glLoadIdentity();
   glMatrixMode( GL_MODELVIEW);

   //readback the texture
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_width, m_height);

   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   UnbindBuffer(FLOAT_BUFFER);
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::ApplyImage( ApplyMode mode, Image2D *image)
{
   int tex0, tex1;
   BindBuffer( FLOAT_BUFFER);

   //set viewport and proj matrix
   glViewport( 0, 0, m_width, m_height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0f, 1.0f, 0.0f, 1.0f);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();


   switch (mode)
   {
      case MODULATE:
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::modulate);
         tex0 = 0;
         tex1 = 1;
         break;
      case ADD:
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::add);
         tex0 = 0;
         tex1 = 1;
         break;
      case SUBTRACT:
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::subtract);
         tex0 = 0;
         tex1 = 1;
         break;
      case REPLACE:
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::replace);
         tex0 = 0;
         tex1 = 1;
         break;
      case NOOP:
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::replace);
         tex0 = 1;
         tex1 = 0;
         break;
   
   };

   //turn on textures and bind
   glActiveTextureARB( GL_TEXTURE0_ARB + tex0);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable(GL_TEXTURE_2D);
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();
   glScalef( (float)m_width/(float)m_pow2Width, (float)m_height/(float)m_pow2Height, 1.0f);
   
   glActiveTextureARB( GL_TEXTURE0_ARB + tex1);
   glBindTexture( GL_TEXTURE_2D, image->m_fTexHandle);
   glEnable(GL_TEXTURE_2D);
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();
   glScalef( (float)image->m_width/(float)image->m_pow2Width, (float)image->m_height/(float)image->m_pow2Height, 1.0f);

   glMatrixMode(GL_MODELVIEW);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);

   glBegin(GL_QUADS);
     glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
     glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.0f);
     glVertex3f( 0.0f, 0.0f, -0.5f);
     glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
     glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.0f);
     glVertex3f( 1.0f, 0.0f, -0.5f);
     glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
     glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 1.0f);
     glVertex3f( 1.0f, 1.0f, -0.5f);
     glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
     glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 1.0f);
     glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, 0);

   glActiveTextureARB( GL_TEXTURE0_ARB + tex0);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_width, m_height);
   
   

   //turn off textures
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable(GL_TEXTURE_2D);
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();
   
   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable(GL_TEXTURE_2D);
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity();

   glMatrixMode(GL_MODELVIEW);
   glActiveTextureARB( GL_TEXTURE0_ARB);

   UnbindBuffer(FLOAT_BUFFER);
}





//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::Copy( Image2D *dest)
{
   BindBuffer( FLOAT_BUFFER);

   //Configure transform
   glViewport( 0, 0, m_pow2Width, m_pow2Height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   //configure the texture
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable( GL_TEXTURE_2D);

   //render the quad
   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   //copy the relevent data
   dest->m_height = m_height;
   dest->m_width = m_width;
   dest->m_pow2Height = m_pow2Height;
   dest->m_pow2Width = m_pow2Width;
   dest->m_wrap = m_wrap;
   dest ->m_type = m_type;

   
   if (!dest->m_fTexHandle)
   {
      glGenTextures( 1, &(dest->m_fTexHandle));

      glBindTexture( GL_TEXTURE_2D, dest->m_fTexHandle);
      SetParameters();
   }

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, dest->m_fTexHandle);
   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, dest->m_pow2Width, dest->m_pow2Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, dest->m_width, dest->m_height);
   
   glDisable(GL_TEXTURE_2D);
   glBindTexture( GL_TEXTURE_2D, 0);

   UnbindBuffer(FLOAT_BUFFER);
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
int Image2D::GetWidth()
{
   return m_width;
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
int Image2D::GetHeight()
{
   return m_height;
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
ImageType Image2D::GetType()
{
   return m_type;
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
bool Image2D::ImageLoad( int width, int height, ImageType type, unsigned char* data)
{
   if ( (width>GetBufferWidth()) || (height>GetBufferHeight()) )
   {
      return false;
   }
   m_width = width;
   m_height = height;
   m_type = type;
   m_pow2Width = Pow2(width);
   m_pow2Height = Pow2(height);

   CheckErr(__LINE__);

   BindBuffer( FLOAT_BUFFER);

   CheckErr(__LINE__);

   if (m_fTexHandle)
   {
      glDeleteTextures( 1, &m_fTexHandle);
   }

   glGenTextures( 1, &m_fTexHandle);

  

   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);

   CheckErr(__LINE__);

   switch (type)
   {
      case MONOCHROME:
         glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, NULL);
         glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, m_width, m_height, GL_LUMINANCE, GL_UNSIGNED_BYTE, data);
         break;
      case LUMINANCE_ALPHA:
         glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, NULL);
         glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, m_width, m_height, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, data);
         break;
      case RGB:
         glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
         glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, m_width, m_height, GL_RGB, GL_UNSIGNED_BYTE, data);
         break;
      case RGBA:
         glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
         glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, m_width, m_height, GL_RGBA, GL_UNSIGNED_BYTE, data);
         break;
   };

   
   SetParameters();

   glBindTexture( GL_TEXTURE_2D, 0);
   
   CheckErr(__LINE__);

   UnbindBuffer(FLOAT_BUFFER);
   return true;
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
bool Image2D::LoadImageFromTexture( int width, int height, ImageType type, unsigned int tex)
{
   if ( (width>GetBufferWidth()) || (height>GetBufferHeight()) )
   {
      return false;
   }
   m_width = width;
   m_height = height;
   m_type = type;
   m_pow2Width = Pow2(width);
   m_pow2Height = Pow2(height);

   BindBuffer( FLOAT_BUFFER);

   if (m_fTexHandle)
   {
      glDeleteTextures( 1, &m_fTexHandle);
   }

   glGenTextures( 1, &m_fTexHandle);

   glBindTexture( GL_TEXTURE_2D,m_fTexHandle);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   SetParameters();

   //Configure transform
   glViewport( 0, 0, m_pow2Width, m_pow2Height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   //configure the texture
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D,tex);
   glEnable( GL_TEXTURE_2D);

   //render the quad
   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   



   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_width, m_height);
   
   glDisable(GL_TEXTURE_2D);
   glBindTexture( GL_TEXTURE_2D, 0);

   UnbindBuffer(FLOAT_BUFFER);
   return true;
}

//
//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::SetParameters()
{
   switch (m_wrap)
   {
      case REPEAT:
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
         break;
      case CLAMP:
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
         break;
      case BORDER:
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER_ARB);
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER_ARB);
         break;
   };
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
}


//
//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::ApplySeperableFilter( ImageFilter *filter)
{
   Convolve2DSeperable *conv = dynamic_cast<Convolve2DSeperable*>(filter);
   GLuint texture;

   conv->Compile();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   if ((conv->HorizontalPasses() > 1) || (conv->VerticalPasses() > 1))
   {
      
      glGenTextures( 1, &texture);
      glBindTexture( GL_TEXTURE_2D, texture);
      glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_RGBA, GL_FLOAT, NULL);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
      
   }

   glEnable(GL_FRAGMENT_PROGRAM_ARB);


   for (int ii=0; ii<conv->HorizontalPasses(); ii++)
   {
      conv->BindHorizontalPass(ii);
      glProgramLocalParameter4fARB( GL_FRAGMENT_PROGRAM_ARB, 0, 1.0f/m_pow2Width, 1.0f/m_pow2Height, 0.0f, 0.0f);
      glBegin(GL_QUADS);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
         glVertex3f( 0.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
         glVertex3f( 1.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
         glVertex3f( 1.0f, 1.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
         glVertex3f( 0.0f, 1.0f, -0.5f);
      glEnd();
      if (ii<conv->HorizontalPasses()-1)
      {
         glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
      }
   }

   

   CheckErr( __LINE__);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
   glActiveTextureARB( GL_TEXTURE1_ARB);

   for (int ii=0; ii<conv->VerticalPasses(); ii++)
   {
      conv->BindVerticalPass(ii);
      glProgramLocalParameter4fARB( GL_FRAGMENT_PROGRAM_ARB, 0, 1.0f/m_pow2Width, 1.0f/m_pow2Height, 0.0f, 0.0f);
      glBegin(GL_QUADS);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
         glVertex3f( 0.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
         glVertex3f( 1.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
         glVertex3f( 1.0f, 1.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
         glVertex3f( 0.0f, 1.0f, -0.5f);
      glEnd();
      if (ii<conv->VerticalPasses()-1)
      {
         glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
      }
   }

   if (conv->Passes() > 1)
   {
      glDeleteTextures(1, &texture);
   }

   glActiveTextureARB( GL_TEXTURE0_ARB);

   glDisable(GL_FRAGMENT_PROGRAM_ARB);


   
}



//
//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::ApplyInseperableFilter( ImageFilter *filter)
{
   Convolve2D *conv = dynamic_cast<Convolve2D*>(filter);
   GLuint texture;

   conv->Compile();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   if (conv->Passes() > 1)
   {
      
      glGenTextures( 1, &texture);
      glBindTexture( GL_TEXTURE_2D, texture);
      glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, m_pow2Width, m_pow2Height, 0, GL_RGBA, GL_FLOAT, NULL);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
      
   }

   glEnable(GL_FRAGMENT_PROGRAM_ARB);


   for (int ii=0; ii<conv->Passes(); ii++)
   {
      conv->BindPass(ii);
      glProgramLocalParameter4fARB( GL_FRAGMENT_PROGRAM_ARB, 0, 1.0f/m_pow2Width, 1.0f/m_pow2Height, 0.0f, 0.0f);
      glBegin(GL_QUADS);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
         glVertex3f( 0.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
         glVertex3f( 1.0f, 0.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
         glVertex3f( 1.0f, 1.0f, -0.5f);
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
         glVertex3f( 0.0f, 1.0f, -0.5f);
      glEnd();
      if (ii<conv->Passes()-1)
      {
         glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
      }
   }

   if (conv->Passes() > 1)
   {
      glDeleteTextures(1, &texture);
   }

   CheckErr( __LINE__);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glDisable(GL_FRAGMENT_PROGRAM_ARB);
   
}
