//
// ATI Manbelbrot
//
//  This demo highlights how to use a floating point off-screen render target with
// ARB_fragment program to handle extremely complex shading calculations. The particular 
// algorithm being run is the Mandelbrot set
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <math.h>

#include <gl/glut.h>
#include "ImageIO.h"
#include "ATIExtensions.h"
#include "3ds.h"
#include "ArcBall.h"
#include "PBuffer.h"
#include "Timer.h"

#ifndef MIN
#define MIN( a, b) ( (a<b) ? a : b)
#endif

//force ANSI compliant for loops
#define for if(0);else for


//number of passes to use
int passes = 4;

//initial window with and height (max coded to 512)
int width = 400;
int height = 400;

//data for the model
data3ds_t model;
float distance = 0.0f;
float pos[3];

//handles to the shaders
GLuint fShad[3];
GLuint vShad[3];

//texture handles
GLuint tempBuffer;
GLuint cmap;

//data file for the object
char mFile[256] = "../models/CrackedQuad2.3ds";


//shader files
//  #0 is for the first pass and sets up the Mandelbrot
//  #1 refines it, and can be chained together
//  #2 applies a color map and performs a phong illumination            
char fsFile0[256] = "mandelFrag0.txt";
char vsFile0[256] = "mandelVert0.txt";
char fsFile1[256] = "mandelFrag1.txt";
char vsFile1[256] = "mandelVert1.txt";
char fsFile2[256] = "mandelFrag2.txt";
char vsFile2[256] = "mandelVert2.txt";

//simple class to handle navigation
ArcBall ball;

//class to deal with PBuffer allocation
PBuffer *buf;

//class to keep track of speed
Timer *time;

//fov used as a zoom factor
float zoom = 60.0f;



//
//  Call-back for mouse movements
//
//
//////////////////////////////////////////////////////////////////////
void trackMotion(int x, int y) {
   if (ball.Track())
   {
      ball.TrackMotion( x, y, glutGet(GLUT_ELAPSED_TIME));
      glutPostRedisplay();
   }
   
}

//
//  Call-back for mouse button events
//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y){
   
   if (button == GLUT_LEFT_BUTTON)
   {
      if (state == GLUT_DOWN)
      {
         ball.StartMotion( x, y, glutGet(GLUT_ELAPSED_TIME));
      }
      else if (state == GLUT_UP)
      {
         ball.StopMotion( x, y, glutGet(GLUT_ELAPSED_TIME));
         
      }
   }
   
   glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void keys( unsigned char c, int x, int y)
{
   switch (c)
   {
      case '\033':
         Free3dsData( &model);
         delete buf;
         delete time;
         exit(0);
         break;

      case '+':
         passes ++;
         break;

      case '-':
         passes --;
         if (passes <2)
         {
            passes = 2;
         }
         break;
   };
   glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void special( int key, int x, int y)
{
   switch (key)
   {
      case GLUT_KEY_UP:
         pos[1] += distance*0.025f;
         break;
      case GLUT_KEY_DOWN:
         pos[1] -= distance*0.025f;
         break;
      case GLUT_KEY_LEFT:
         pos[0] -= distance*0.025f;
         break;
      case GLUT_KEY_RIGHT:
         pos[0] += distance*0.025f;
         break;
      case GLUT_KEY_PAGE_UP:
         if (zoom < 179.5f)
         {
            zoom += 0.5f;
         }
         break;
      case GLUT_KEY_PAGE_DOWN:
         if (zoom > 0.5f)
         {
            zoom -= 0.5f;
         }
         break;
   };
   glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void drawModel()
{
   //
   //Drawn in immediate mode for convenience, for performance, use vertex arrays with VAO
   //
   glBegin( GL_TRIANGLES);
   int vert;
   for (int ii=0; ii<model.meshCount; ii++)
   {
      for (int gg=0; gg<model.meshes[ii].triCount; gg++)
      {
         vert = model.meshes[ii].tris[gg][0];
         if (model.meshes[ii].texCoords)
         {
            glMultiTexCoord2fvARB( GL_TEXTURE0_ARB, model.meshes[ii].texCoords[vert]);
            glMultiTexCoord3fvARB( GL_TEXTURE1_ARB, &(model.meshes[ii].tangentSpace[gg*3][0]));
            glMultiTexCoord3fvARB( GL_TEXTURE2_ARB, &(model.meshes[ii].tangentSpace[gg*3][3]));
         }
         glNormal3fv( model.meshes[ii].norms[gg*3]);
         glVertex3fv( model.meshes[ii].verts[vert]);
         vert = model.meshes[ii].tris[gg][1];
         if (model.meshes[ii].texCoords)
         {
            glMultiTexCoord2fvARB( GL_TEXTURE0_ARB, model.meshes[ii].texCoords[vert]);
            glMultiTexCoord3fvARB( GL_TEXTURE1_ARB, &(model.meshes[ii].tangentSpace[gg*3+1][0]));
            glMultiTexCoord3fvARB( GL_TEXTURE2_ARB, &(model.meshes[ii].tangentSpace[gg*3+1][3]));
         }
         glNormal3fv( model.meshes[ii].norms[gg*3+1]);
         glVertex3fv( model.meshes[ii].verts[vert]);
         vert = model.meshes[ii].tris[gg][2];
         if (model.meshes[ii].texCoords)
         {
            glMultiTexCoord2fvARB( GL_TEXTURE0_ARB, model.meshes[ii].texCoords[vert]);
            glMultiTexCoord3fvARB( GL_TEXTURE1_ARB, &(model.meshes[ii].tangentSpace[gg*3+2][0]));
            glMultiTexCoord3fvARB( GL_TEXTURE2_ARB, &(model.meshes[ii].tangentSpace[gg*3+2][3]));
         }
         glNormal3fv( model.meshes[ii].norms[gg*3+2]);
         glVertex3fv( model.meshes[ii].verts[vert]);
      }
       
   }
   glEnd();
}

//
// drawToTemp
//
//   This is the function that does all the work of the multipassing
//
//////////////////////////////////////////////////////////////////////
void drawToTemp()
{
   //bind the PBuffer as the render target
   buf->Bind();

   glViewport( 0, 0, width, height);

   glClearColor( 0.5f, 0.5f, 0.5f, 0.0f);
   
   glDrawBuffer(  GL_FRONT_AND_BACK);
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluPerspective( zoom, (float)width/(float)height, distance/200.0f, distance*2.0f);

   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef( -pos[0], -pos[1], -pos[2]);
   
   
   ball.Update();
   glMultMatrixf(ball.GetMatrix());
   glRotatef( 90.0f, 0.0f, 1.0f, 0.0f);
   

   glEnable( GL_VERTEX_PROGRAM_ARB);
   glEnable( GL_FRAGMENT_PROGRAM_ARB);

   //leave these two textures available for all shaders
   glBindTexture( GL_TEXTURE_2D, tempBuffer);
   glActiveTexture( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, cmap);
   glActiveTexture( GL_TEXTURE0_ARB);


   for (int pass=0; pass<passes; pass++)
   {
      //select the correct buffers to use as render targets and textures
      // this prevents swaps, which could add copy overhead
      // rendering and texturing trade buffers each iteration
      glDrawBuffer( (pass&0x1) ? GL_BACK_LEFT : GL_FRONT_LEFT);
      buf->BindAsTexture((pass&0x1) ? WGL_FRONT_LEFT_ARB : WGL_BACK_LEFT_ARB);

      //configure the right shaders for this pass
      if (pass == 0)
      {
         //bind initial shader
         glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vShad[0]);
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, fShad[0]);
      }
      else if (pass == passes-1)
      {
         //bind final shader
         glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vShad[2]);
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, fShad[2]);
      }
      else
      {
         glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vShad[1]);
         glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, fShad[1]);
      }

      //draw the geometry
      drawModel();

      //release the texture, so it can be used to rneder to
      buf->ReleaseTexture( (pass&0x1) ? WGL_FRONT_LEFT_ARB : WGL_BACK_LEFT_ARB);
   }

   glDisable( GL_VERTEX_PROGRAM_ARB);
   glDisable( GL_FRAGMENT_PROGRAM_ARB);

   //return to drawing to the window
   buf->Unbind();
}

//
//
//////////////////////////////////////////////////////////////////////
void draw()
{
   char buffer[256];
   
   //draw the scene to the float buffer
   drawToTemp();
   
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   glOrtho( 0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 1.0f);

   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   //bind the correct version of the PBuffer
   glBindTexture( GL_TEXTURE_2D, tempBuffer);
   buf->BindAsTexture( ((passes)&0x1) ? WGL_FRONT_LEFT_ARB : WGL_BACK_LEFT_ARB);
   glEnable( GL_TEXTURE_2D);

   glColor3f( 1.0f, 1.0f, 1.0f);

   glBegin( GL_QUADS);
   glTexCoord2f( 0.0f, 0.0f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glTexCoord2f( width/512.0f, 0.0f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glTexCoord2f( width/512.0f, height/512.0f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glTexCoord2f( 0.0f, height/512.0f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   //release it so it can be rendered to
   buf->ReleaseTexture( ((passes)&0x1) ? WGL_FRONT_LEFT_ARB : WGL_BACK_LEFT_ARB);

   //draw text to the screen
   glDisable( GL_TEXTURE_2D);
   
   glRasterPos2f( 0.75f, 0.95f);
   sprintf( buffer, "FPS: %3.2f", time->GetAvgHz());
   for (char* c=buffer; *c != '\0'; c++)
   {
      glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, *c);
   }
   

   glRasterPos2f( 0.01f, 0.01f);
   sprintf( buffer, "Num Passes: %d", passes);
   for (char* c=buffer; *c != '\0'; c++)
   {
      glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, *c);
   }

   time->UpdateTimer();
   glutSwapBuffers();
}

//
//
//////////////////////////////////////////////////////////////////////
void reshape( int w, int h)
{
   width = w;
   height = h;

   glViewport( 0, 0, width, height);
   ball.SetWidth(width);
   ball.SetHeight(height);

   //clamp the window size
   if ((w>512) || (h>512))
   {
      w = MIN(w, 512);
      h = MIN(h, 512);
      glutReshapeWindow( w, h);
   }
}

//
// Simple function to loaf a fragment program
//
////////////////////////////////////////////////////////////////////////////////
GLuint loadFragmentProgram( char* fn)
{
   FILE *fp;
   GLubyte *buf;
   int length;
   bool ret = true;
   GLuint rp;

   if (!(fp = fopen(fn,"rb")))
   {
      return false;
   }

   fseek(fp, 0, SEEK_END);
   length = ftell(fp);
   fseek(fp, 0, SEEK_SET);

   buf = new GLubyte[length+1];

   fread( buf, 1, length, fp);
   buf[length] = '\0'; // make it a regular C string so str* functions work

   glGenProgramsARB( 1, &rp);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, rp);

   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, length, buf);

   if (glGetError() != 0)
   {
      int position;
      printf("%s\n", glGetString(GL_PROGRAM_ERROR_STRING_ARB));
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &position);
      printf(" @%d\n", position);
      ret = false;
   }

   fclose(fp);

   delete []buf;
   return ret ? rp : 0;
}

//
// Simple function to load a vertex program
//
////////////////////////////////////////////////////////////////////////////////
GLuint loadVertexProgram( char* fn)
{
   FILE *fp;
   GLubyte *buf;
   int length;
   bool ret = true;
   GLuint vp;

   if (!(fp = fopen(fn,"rb")))
   {
      return false;
   }

   fseek(fp, 0, SEEK_END);
   length = ftell(fp);
   fseek(fp, 0, SEEK_SET);

   buf = new GLubyte[length+1];

   fread( buf, 1, length, fp);
   buf[length] = '\0'; // make it a regular C string so str* functions work

   glGenProgramsARB( 1, &vp);
   glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vp);

   glProgramStringARB( GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, length, buf);

   if (glGetError() != 0)
   {
      int lineno;
      printf("%s\n", glGetString(GL_PROGRAM_ERROR_STRING_ARB));
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &lineno);
      printf(" @%d\n", lineno);
      ret = false;
   }

   fclose(fp);

   delete []buf;
   return ret ? vp :0;
}

//
//
//////////////////////////////////////////////////////////////////////
void init()
{
   GLubyte colorRamp[] =
   {
      0xff, 0xff, 0xff, 0xff,
      0xff, 0x00, 0x00, 0xff,
      0x00, 0xff, 0x00, 0xff,
      0x00, 0x00, 0xff, 0xff
   };

   //make the model file work
   Read3dsFile( mFile, &model);
   Calculate3dsNormals( &model);
   Calculate3dsTangentSpace( &model);
   Calculate3dsBoundingBox( &model);

   float x,y,z;

   x = model.max[0] - model.min[0];
   y = model.max[1] - model.min[1];
   z = model.max[2] - model.min[2];
   distance = sqrtf( x*x + y*y + z*z); 
   pos[0] = 0.0f;
   pos[1] = 0.0f;
   pos[2] = distance;

   buf->Bind();

   

   vShad[0] = loadVertexProgram(vsFile0);
   fShad[0] = loadFragmentProgram(fsFile0);
   vShad[1] = loadVertexProgram(vsFile1);
   fShad[1] = loadFragmentProgram(fsFile1);
   vShad[2] = loadVertexProgram(vsFile2);
   fShad[2] = loadFragmentProgram(fsFile2);

   glGenTextures( 1, &tempBuffer);
   glBindTexture( GL_TEXTURE_2D, tempBuffer);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);

   glGenTextures( 1, &cmap);
   glBindTexture( GL_TEXTURE_2D, cmap);
   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, 4, 1, 0, GL_RGBA, GL_UNSIGNED_BYTE, colorRamp);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);

   
   glDepthFunc( GL_LEQUAL);
   glEnable( GL_DEPTH_TEST);
   buf->Unbind();

}



//
//
//////////////////////////////////////////////////////////////////////
int main( int argc, char **argv)
{
   glutInit( &argc, argv);
   glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
   glutInitWindowSize( width, height);
   ball.SetWidth(width);
   ball.SetHeight(height);

   glutCreateWindow("ATI Mandelbrot");

   printf("ATI Mandelbrot\n");
   printf("  '+'       - increase number of iterations\n");
   printf("  '-'       - decrease number of iterations\n");
   printf("  [UP]      - translate up\n");
   printf("  [DOWN]    - translate down\n");
   printf("  [LEFT]    - translate left\n");
   printf("  [RIGHT]   - translate right\n");
   printf("  [PAGEUP]  - Zoom out\n");
   printf("  [PAGEDOWN]- Zoom in\n");
   printf("  [ESC]     - quit\n");

   glutDisplayFunc( draw);
   glutIdleFunc( draw);
   glutKeyboardFunc( keys);
   glutSpecialFunc( special);
   glutReshapeFunc( reshape);
   glutMouseFunc(mouse);
   glutMotionFunc(trackMotion);

   SetupGL1_3();
   SetupARBExtensions();
   SetupWGLExtensions();

   buf = new PBuffer( 512, 512, 128, true, true, true, false, true, true);
   time = new Timer();

   if (!buf->IsValid())
   {
      printf( "Failed to alloc PBuffer\n");
      exit(1);
   }

   glClearColor( 0.5f, 0.5f, 0.5f, 0.0f);

   init();

   glutMainLoop();
   return 0;
}