//
// Simple_Shader.cpp
//
//   Program for interactively displaying shaders on a model file. 
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <io.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/locking.h>
#include <fcntl.h>
#include <share.h>
#include <math.h>

#include <gl/glut.h>
#include "ImageIO.h"
#include "ATIExtensions.h"
#include "3ds.h"
#include "ArcBall.h"
#include "Timer.h"

//force ANSI compliant for loops
#define for if(0);else for

#define MAX_TEXTURES 8

int width = 400;
int height = 400;


float distance = 0.0f;

//state for shaders, textures etc
GLuint fShad;
GLuint vShad;
char vsFile[256] = "Shaders/BrickVert.txt";
char fsFile[256] = "Shaders/BrickFrag.txt";
char mFile[256] = "logo.3ds";
char tFile[MAX_TEXTURES][256] = { "", "", "", "", "", "", "", ""};
GLuint tID[MAX_TEXTURES];
Image tImages[MAX_TEXTURES] = { 0, 0, 0, 0, 0, 0, 0, 0};
data3ds_t model;

//utilities for timing and navigation
ArcBall ball;
Timer *time = NULL;

//drawing options
int drawMode = 0;
bool cull = true;
bool lineSmooth = false;

//display options
int showStats = 1;

//options for contolling shader reloading
bool reload = true;
int frame = 0;

//this is used as the FOV
float zoom = 60.0f;

void draw();
void reloadShaders( int val);



//
//  Call-back for mouse movements
//
//
//////////////////////////////////////////////////////////////////////
void trackMotion(int x, int y) {
   if (ball.Track())
   {
      ball.TrackMotion( x, y, glutGet(GLUT_ELAPSED_TIME));
      glutPostRedisplay();
   }
   
}

//
//  Call-back for mouse button events
//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y){
   
   if (button == GLUT_LEFT_BUTTON)
   {
      if (state == GLUT_DOWN)
      {
         ball.StartMotion( x, y, glutGet(GLUT_ELAPSED_TIME));
      }
      else if (state == GLUT_UP)
      {
         ball.StopMotion( x, y, glutGet(GLUT_ELAPSED_TIME));
      }
   }
   
   glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void keys( unsigned char c, int x, int y)
{
   switch (c)
   {
      case '\033':
         Free3dsData( &model);
         delete time;
         exit(0);
         break;

      case 'r':
         reload =! reload;
         if (reload)
         {
            glutTimerFunc( 100, reloadShaders, 0);
         }
         break;

      case 's':
         showStats = (showStats+1)%3;
         break;

      case 'd':
         drawMode++;
         drawMode %= 4;
         break;

      case 'c':
         cull = !cull;
         break;

      case 'a':
         lineSmooth = !lineSmooth;
         break;

      case '+':
         if (zoom > 0.5f)
            zoom -= 0.5f;
         break;

      case '-':
         if (zoom < 179.5f)
            zoom += 0.5f;
         break;
   };

   glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void drawModel()
{
   //
   //Drawn in immediate mode for convenience, for performance, use vertex arrays with VAO
   //
   glBegin( GL_TRIANGLES);
   int vert;
   for (int ii=0; ii<model.meshCount; ii++)
   {
      for (int gg=0; gg<model.meshes[ii].triCount; gg++)
      {
         vert = model.meshes[ii].tris[gg][0];
         if (model.meshes[ii].texCoords)
         {
            glMultiTexCoord2fvARB( GL_TEXTURE0_ARB, model.meshes[ii].texCoords[vert]);
            glMultiTexCoord3fvARB( GL_TEXTURE1_ARB, &(model.meshes[ii].tangentSpace[gg*3][0]));
            glMultiTexCoord3fvARB( GL_TEXTURE2_ARB, &(model.meshes[ii].tangentSpace[gg*3][3]));
         }
         glNormal3fv( model.meshes[ii].norms[gg*3]);
         glVertex3fv( model.meshes[ii].verts[vert]);
         vert = model.meshes[ii].tris[gg][1];
         if (model.meshes[ii].texCoords)
         {
            glMultiTexCoord2fvARB( GL_TEXTURE0_ARB, model.meshes[ii].texCoords[vert]);
            glMultiTexCoord3fvARB( GL_TEXTURE1_ARB, &(model.meshes[ii].tangentSpace[gg*3+1][0]));
            glMultiTexCoord3fvARB( GL_TEXTURE2_ARB, &(model.meshes[ii].tangentSpace[gg*3+1][3]));
         }
         glNormal3fv( model.meshes[ii].norms[gg*3+1]);
         glVertex3fv( model.meshes[ii].verts[vert]);
         vert = model.meshes[ii].tris[gg][2];
         if (model.meshes[ii].texCoords)
         {
            glMultiTexCoord2fvARB( GL_TEXTURE0_ARB, model.meshes[ii].texCoords[vert]);
            glMultiTexCoord3fvARB( GL_TEXTURE1_ARB, &(model.meshes[ii].tangentSpace[gg*3+2][0]));
            glMultiTexCoord3fvARB( GL_TEXTURE2_ARB, &(model.meshes[ii].tangentSpace[gg*3+2][3]));
         }
         glNormal3fv( model.meshes[ii].norms[gg*3+2]);
         glVertex3fv( model.meshes[ii].verts[vert]);
      }
       
   }
   glEnd();
}

//
//
//////////////////////////////////////////////////////////////////////
void draw()
{
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluPerspective( zoom, (float)width/(float)height, distance/10.0f, distance*2.0f);

   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef( 0.0f, 0.0f, -distance);


   ball.Update();
   
   glMultMatrixf(ball.GetMatrix());
   
   
   //setup textures, don't bother to enable as fragment program overrides them anyway
   for (int ii=0; ii<MAX_TEXTURES; ii++)
   {
      if (tID[ii])
      {
         glActiveTextureARB( GL_TEXTURE0_ARB + ii);
         if ( imgGetDepth(tImages[ii]) <2)
         {
            glBindTexture( GL_TEXTURE_2D, tID[ii]);
         }
         else
         {
            glBindTexture( GL_TEXTURE_3D_EXT, tID[ii]);
         }
      }
   }
   glActiveTextureARB( GL_TEXTURE0_ARB );
   

   //enable vertex and fragment program
   glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vShad);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, fShad);

   //configure the drawing options
   if (lineSmooth)
   {
      glEnable(GL_LINE_SMOOTH);
      glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glLineWidth(1.0f);
   }
   else
   {
      glBlendFunc( GL_ONE, GL_ZERO);
      glLineWidth(1.0f);
   }

   if (cull)
   {
      glCullFace(GL_BACK);
      glEnable(GL_CULL_FACE);
   }

   switch (drawMode)
   {
      case 0:
         //standard filled drawing
         glEnable( GL_VERTEX_PROGRAM_ARB);
         glEnable( GL_FRAGMENT_PROGRAM_ARB);

         drawModel();
         glDisable( GL_VERTEX_PROGRAM_ARB);
         glDisable( GL_FRAGMENT_PROGRAM_ARB);

         break;
      case 1:
         //regular wireframe
         glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
         glEnable( GL_VERTEX_PROGRAM_ARB);
         glEnable( GL_FRAGMENT_PROGRAM_ARB);
         glEnable( GL_BLEND);

         drawModel();
         glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
         glDisable( GL_BLEND);
         glDisable( GL_VERTEX_PROGRAM_ARB);
         glDisable( GL_FRAGMENT_PROGRAM_ARB);
         break;
      case 2:
         //hidden line
         glEnable( GL_VERTEX_PROGRAM_ARB);
         glEnable( GL_FRAGMENT_PROGRAM_ARB);
         glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
         glPolygonOffset( 1.0f, 1.0f);
         glEnable( GL_POLYGON_OFFSET_FILL);

         drawModel();
         glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
         glDisable( GL_POLYGON_OFFSET_FILL);

         glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
         glEnable( GL_BLEND);
         drawModel();
         glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
         glDisable( GL_BLEND);
         glDisable( GL_VERTEX_PROGRAM_ARB);
         glDisable( GL_FRAGMENT_PROGRAM_ARB);
         break;
      case 3:
         //overlaid wireframe
         glEnable( GL_VERTEX_PROGRAM_ARB);
         glEnable( GL_FRAGMENT_PROGRAM_ARB);
         glPolygonOffset( 1.0f, 1.0f);
         glEnable( GL_POLYGON_OFFSET_FILL);

         drawModel();
         glDisable( GL_POLYGON_OFFSET_FILL);
         glDisable( GL_VERTEX_PROGRAM_ARB);
         glDisable( GL_FRAGMENT_PROGRAM_ARB);

         glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
         glEnable( GL_BLEND);
         drawModel();
         glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
         glDisable( GL_BLEND);
         break;
   };

   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_CULL_FACE);
   
   glActiveTextureARB( GL_TEXTURE0_ARB);

   

   
   time->UpdateTimer();
   if (showStats)
   {
      float x = 5.0f, y= height-20.0f;
      char buffer[256];

      //configure the transforms for 2D
      glMatrixMode( GL_PROJECTION);
      glLoadIdentity();
      gluOrtho2D( 0.0f, (float)width, 0.0f, (float)height);
      glMatrixMode( GL_MODELVIEW);
      glLoadIdentity();

      //draw the frame counter
      glRasterPos2f( x, y);
      sprintf( buffer, "FPS: %3.2f", time->GetAvgHz());
      glColor3f( 1.0f, 1.0f, 1.0f);
      for (char *c=buffer; (*c); c++)
      {
         glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, *c);
      }
      
      //show extended stats
      if (showStats > 1)
      {
         //show the draw mode
         y -= 20.0f;
         glRasterPos2f( x, y);
         sprintf( buffer, "Draw mode: %d", drawMode);
         glColor3f( 1.0f, 1.0f, 1.0f);
         for (char *c=buffer; (*c); c++)
         {
            glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, *c);
         }

         //show culling on/off
         y -= 12.0f;
         glRasterPos2f( x, y);
         sprintf( buffer, "Culling: %s", ((cull) ? "On" : "Off"));
         glColor3f( 1.0f, 1.0f, 1.0f);
         for (char *c=buffer; (*c); c++)
         {
            glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, *c);
         }

         //show aa on/off
         y -= 12.0f;
         glRasterPos2f( x, y);
         sprintf( buffer, "AA wireframe: %s", ((lineSmooth) ? "On" : "Off"));
         glColor3f( 1.0f, 1.0f, 1.0f);
         for (char *c=buffer; (*c); c++)
         {
            glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, *c);
         }
      }
   }
   glutSwapBuffers();

   
}

//
//
//////////////////////////////////////////////////////////////////////
void reshape( int w, int h)
{
   width = w;
   height = h;

   glViewport( 0, 0, width, height);
   ball.SetWidth(width);
   ball.SetHeight(height);
}

//
//  The follwing ifdef'd sections are meant to deal with a file sharing issue. 
// By default the program will try to autoreload its shaders every 4th frame.
// Some editors look for file access and don't like this, so a path that forces
// a manual reload is also available.
//

//
// Simple function to load a fragment program
//  This version always attempts to load, returning 0 on failure.
////////////////////////////////////////////////////////////////////////////////
GLuint loadFragmentProgram( char* fn)
{
   FILE *fp;
   GLubyte *buf;
   int length;
   bool ret = true;
   GLuint rp;

   if (!(fp = fopen(fn,"rb")))
   {
      return false;
   }

   printf("\nLoading fragment program: '%s'\n", fn);

   fseek(fp, 0, SEEK_END);
   length = ftell(fp);
   fseek(fp, 0, SEEK_SET);

   buf = new GLubyte[length+1];

   fread( buf, 1, length, fp);
   buf[length] = '\0'; // make it a regular C string so str* functions work

   glGenProgramsARB( 1, &rp);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, rp);

   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, length, buf);

   if (glGetError() != 0)
   {
      int position;
      printf("%s\n", glGetString(GL_PROGRAM_ERROR_STRING_ARB));
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &position);
      printf(" @%d\n", position);
      ret = false;
   }
   else
   {
      printf("  Load succeeded\n");
   }

   fclose(fp);

   delete []buf;
   return ret ? rp : 0;
}

//
// Simple function to load a vertex program
//  This function always attempts to load, returning 0 on failure.
////////////////////////////////////////////////////////////////////////////////
GLuint loadVertexProgram( char* fn)
{
   FILE *fp;
   GLubyte *buf;
   int length;
   bool ret = true;
   GLuint vp;

   if (!(fp = fopen(fn,"rb")))
   {
      return false;
   }

   printf("\nLoading vertex program: '%s'\n", fn);

   fseek(fp, 0, SEEK_END);
   length = ftell(fp);
   fseek(fp, 0, SEEK_SET);

   buf = new GLubyte[length+1];

   fread( buf, 1, length, fp);
   buf[length] = '\0'; // make it a regular C string so str* functions work

   glGenProgramsARB( 1, &vp);
   glBindProgramARB( GL_VERTEX_PROGRAM_ARB, vp);

   glProgramStringARB( GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, length, buf);

   if (glGetError() != 0)
   {
      int lineno;
      printf("%s\n", glGetString(GL_PROGRAM_ERROR_STRING_ARB));
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &lineno);
      printf(" @%d\n", lineno);
      ret = false;
   }
   else
   {
      printf("  Load succeeded\n");
   }

   fclose(fp);

   delete []buf;
   return ret ? vp :0;
}



//
//
//////////////////////////////////////////////////////////////////////
void init()
{
   //
   //load the model file
   //
   Read3dsFile( mFile, &model);
   Calculate3dsNormals( &model);
   Calculate3dsTangentSpace( &model);
   Calculate3dsBoundingBox( &model);

   if (model.triCount == 0)
   {
      printf("Error: Model file '%s' is either missing or lacks triangles\n", mFile);
   }

   float x,y,z;

   x = model.max[0] - model.min[0];
   y = model.max[1] - model.min[1];
   z = model.max[2] - model.min[2];
   distance = sqrtf( x*x + y*y + z*z); 

   //
   //Load the shader files
   //
   reloadShaders(0);
   /*
   vShad = loadVertexProgram(vsFile);
   fShad = loadFragmentProgram(fsFile);
   */

   //
   //Load the textures
   //
   for (int ii=0; ii<MAX_TEXTURES; ii++)
   {
      if (tFile[ii][0] != '\0')
      {
         tImages[ii] = imgCreateImage();
         if (imgLoadImage( tImages[ii], tFile[ii]))
         {
            //failed
            printf("Error: Unable to open image file %s\n", tFile[ii]);
            imgDestroyImage( tImages[ii]);
            tImages[ii] = 0;
            tID[ii] = 0;
         }
         else
         {
            glGenTextures( 1, &tID[ii]);
            if ( imgGetDepth(tImages[ii]) < 2)
            {
               glBindTexture( GL_TEXTURE_2D, tID[ii]);
               glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
               imgLoadImageAsTexture( tImages[ii]);
               glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
               glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
               glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
               glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            }
            else
            {
               glBindTexture( GL_TEXTURE_3D_EXT, tID[ii]);
               imgLoadImageAsTexture( tImages[ii]);
               glTexParameteri( GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_REPEAT);
               glTexParameteri( GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_REPEAT);
               glTexParameteri( GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_REPEAT);
               glTexParameteri( GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
               glTexParameteri( GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            }
         }
      }
      else
      {
         tID[ii] = 0;
      }
   }

   glDepthFunc( GL_LEQUAL);
   glEnable( GL_DEPTH_TEST);
   time = new Timer();
}

//
//
//////////////////////////////////////////////////////////////////////
void reloadShaders( int val)
{
   static unsigned int vTime = 0;
   static unsigned int fTime = 0;
   struct _stat fstats;
   bool frag = false;
   bool vert = false;
   GLuint tvShad = 0, tfShad = 0;

   _stat( vsFile, &fstats);
   if (fstats.st_mtime > vTime)
   {
      vTime = fstats.st_mtime;
      vert = true;
   }

   _stat( fsFile, &fstats);
   if (fstats.st_mtime > fTime)
   {
      fTime = fstats.st_mtime;
      frag = true;
   }
   
   if (vert)
   {
      tvShad = loadVertexProgram(vsFile);
   }

   if (frag)
   {
      tfShad = loadFragmentProgram(fsFile);
   }
   
   if (tvShad)
   {
      glDeleteProgramsARB( 1, &vShad);
      vShad = tvShad;
   }

   
   if (tfShad)
   {
      glDeleteProgramsARB( 1, &fShad);
      fShad = tfShad;
   }

   if (reload)
   {
      glutTimerFunc( 100, reloadShaders, 0);
   }
}

//
//
//////////////////////////////////////////////////////////////////////
void ProcessCommandLine( int argc, char **argv)
{

   for (int ii=0; ii<argc; ii++)
   {
      if (!strcmp( argv[ii], "-v"))
      {
         strcpy( vsFile, argv[ii+1]);
         ii++;
      }
      else if (!strcmp( argv[ii], "-f"))
      {
         strcpy( fsFile, argv[ii+1]);
         ii++;
      }
      else if (!strcmp( argv[ii], "-m"))
      {
         strcpy( mFile, argv[ii+1]);
         ii++;
      }
      else if (!strncmp( argv[ii], "-t", 2))
      {
         int which = atoi( &argv[ii][2]);
         if (which >= MAX_TEXTURES)
         {
            continue;
         }
         strcpy( tFile[which], argv[ii+1]);
         ii++;
      }
   }
}

//
//
//////////////////////////////////////////////////////////////////////
int main( int argc, char **argv)
{
   glutInit( &argc, argv);
   glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
   glutInitWindowSize( width, height);
   ball.SetWidth(width);
   ball.SetHeight(height);

   glutCreateWindow("Simple Shader- ARB Fragment and Vertex Program loader");

   printf("Simple Shader\n");
   printf("  [ESC] - quit program\n");
   printf("  'r'   - toggle auto reload of shaders\n");
   printf("  's'   - change stats mode\n");
   printf("  'd'   - change draw mode\n");
   printf("  'c'   - toggle back face culling\n");
   printf("  '+'   - zoom in\n");
   printf("  '-'   - zoom out\n");
   printf("  'a'   - toggle antialiasing\n\n");

   glutDisplayFunc( draw);
   glutIdleFunc( draw);
   glutKeyboardFunc( keys);
   glutReshapeFunc( reshape);
   glutMouseFunc(mouse);
   glutMotionFunc(trackMotion);
   

   SetupGL1_2();
   SetupARBExtensions();

   if ( (!GL_ARB_fragment_program_Flag) || (!GL_ARB_vertex_program_Flag))
   {
      printf("Error, this program requires ARB_fragment_program and ARB_vertex_program\n");
      return 1;
   }

   ProcessCommandLine( argc, argv);

   init();

   glutMainLoop();
   return 0;
}