/*
	Emil Persson
	epersson@ati.com
	ATI Technologies Inc.

	A class for simplifying the handling of common stuff such as texture objects,
	shaders and blend modes.
*/

#include "Renderer.h"
#include "../include/ImageIO.h"
#include "noise.h"
#include <stdio.h>
#include <math.h>

unsigned int packf(const float r, const float g, const float b, const float a){
	union {
		unsigned char rgba[4];
		unsigned int final;
	} pixel;

	pixel.rgba[0] = (r < 0)? 0 : (r > 1)? 255 : (unsigned char) (r * 255.0f);
	pixel.rgba[1] = (g < 0)? 0 : (g > 1)? 255 : (unsigned char) (g * 255.0f);
	pixel.rgba[2] = (b < 0)? 0 : (b > 1)? 255 : (unsigned char) (b * 255.0f);
	pixel.rgba[3] = (a < 0)? 0 : (a > 1)? 255 : (unsigned char) (a * 255.0f);

	return pixel.final;
}


Renderer::Renderer(){
	defaults();

	hdc = wglGetCurrentDC();
	hglrc = wglGetCurrentContext();
}

Renderer::~Renderer(){
	for (unsigned int i = 0; i < textures.getSize(); i++){
		if (textures[i].isRenderTarget){
			wglDeleteContext(textures[i].hPrc);
			wglReleasePbufferDCARB(textures[i].hPBuffer, textures[i].hPdc);
			wglDestroyPbufferARB(textures[i].hPBuffer);
		}

		glDeleteTextures(1, &textures[i].textureID);
	}

}

void Renderer::defaults(){
	for (unsigned int i = 0; i < MAX_TMU; i++){
		currentTextures[i] = TEXTURE_NONE;
	}
	nCurrentTextures = 0;
	currentBlendEnabled = false;
	currentShader = SHADER_NONE;

	reset();
}

void Renderer::reset(){
	for (unsigned int i = 0; i < MAX_TMU; i++){
		selectedTextures[i] = TEXTURE_NONE;
	}
	nSelectedTextures = 0;
	selectedBlendEnabled = false;
	selectedShader = SHADER_NONE;
}

void Renderer::apply(){
	unsigned int nChange = max(nSelectedTextures, nCurrentTextures);

	// Apply texture changes
	for (unsigned int i = 0; i < nChange; i++){
		TextureID currTex = currentTextures[i];
		TextureID selTex = selectedTextures[i];
		
		if (selTex != currTex){
			glActiveTextureARB(GL_TEXTURE0_ARB + i);

			if (currTex != TEXTURE_NONE && textures[currTex].isRenderTarget)
				wglReleaseTexImageARB(textures[currTex].hPBuffer, WGL_FRONT_LEFT_ARB);

			if (selTex == TEXTURE_NONE){
				glDisable(textures[currTex].target);
			} else {
				if (currTex == TEXTURE_NONE){
					glEnable(textures[selTex].target);
				} else {
					if (textures[selTex].target != textures[currTex].target){
						glDisable(textures[currTex].target);
						glEnable(textures[selTex].target);
					}
				}
				
				glBindTexture(textures[selTex].target, textures[selTex].textureID);
				if (textures[selTex].isRenderTarget)
					wglBindTexImageARB(textures[selTex].hPBuffer, WGL_FRONT_LEFT_ARB);
			}
			currentTextures[i] = selTex;
		}
	}

	// Apply changes to vertex and fragment shaders
	if (selectedShader != currentShader){
		if (selectedShader == SHADER_NONE){
			glUseProgramObjectGL2(0);
		} else {
			glUseProgramObjectGL2(shaders[selectedShader].programObj);
		}
		currentShader = selectedShader;
	}

	// Apply changes to blend equation
	if (selectedBlendEnabled){
		if (selectedSrcFactor != currentSrcFactor || selectedDstFactor != currentDstFactor){
			glBlendFunc(selectedSrcFactor, selectedDstFactor);
			currentSrcFactor = selectedSrcFactor;
			currentDstFactor = selectedDstFactor;
		}
		if (!currentBlendEnabled){
			glEnable(GL_BLEND);
			currentBlendEnabled = true;
		}
	} else {
		if (currentBlendEnabled){
			glDisable(GL_BLEND);
			currentBlendEnabled = false;
		}
	}

	nCurrentTextures = nSelectedTextures;

	reset();
}

#define MCHAR4(a, b, c, d) (a | (b << 8) | (c << 16) | (d << 24))

TextureID Renderer::addTexture(const char *fileName){
	Image img = imgCreateImage();
	Texture tex;
	tex.isRenderTarget = false;

	char *ext = strrchr(fileName, '.');
	if (strcmp(ext, ".dds") == 0){
		struct {
			unsigned int  ddsIdentifier;
			unsigned char junk[8];
			unsigned int  height;
			unsigned int  width;
			unsigned char junk2[8];
			unsigned int  nMipMaps;
			unsigned char junk3[52];
			unsigned int  fourCC;
			unsigned char junk4[40];
		} header;

		FILE *file;
		if ((file = fopen(fileName, "rb")) == NULL) return TEXTURE_NONE;
		
		fread(&header, sizeof(header), 1, file);
		if (header.ddsIdentifier != MCHAR4('D', 'D', 'S', ' ')){
			fclose(file);
			return TEXTURE_NONE;
		}

		GLuint format;
		switch(header.fourCC){
		case MCHAR4('D', 'X', 'T', '1'):
			format = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
			break;
		case MCHAR4('D', 'X', 'T', '3'):
			format = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
			break;
		case MCHAR4('D', 'X', 'T', '5'):
			format = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
			break;
		default:
			fclose(file);
			return TEXTURE_NONE;
		}
		
		if (header.nMipMaps == 0) header.nMipMaps = 1;
		int w = header.width;
		int h = header.height;

		glGenTextures(1, &tex.textureID);
		tex.target = GL_TEXTURE_2D;
		glBindTexture(tex.target, tex.textureID);

		unsigned char *pixels = new unsigned char[w * h];
		unsigned int mipMapLevel = 0;
		for (unsigned int i = 0; i < header.nMipMaps; i++){
			int size = ((w + 3) >> 2) * ((h + 3) >> 2) * ((format == GL_COMPRESSED_RGB_S3TC_DXT1_EXT)? 8 : 16);

			fread(pixels, 1, size, file);
			glCompressedTexImage2DARB(tex.target, mipMapLevel, format, w, h, 0, size, pixels);

			if (w > 1) w >>= 1;
			if (h > 1) h >>= 1;
			mipMapLevel++;
		}

		fclose(file);

		delete pixels;


		glTexParameteri(tex.target, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(tex.target, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, header.nMipMaps > 1? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
		glTexParameteri(tex.target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			
		return textures.add(tex);
	} else {
		if (imgLoadImage(img, (char *) fileName)){
			imgDestroyImage(img);
			return TEXTURE_NONE;
		} else {
			glGenTextures(1, &tex.textureID);
			tex.target = GL_TEXTURE_2D;
			glBindTexture(tex.target, tex.textureID);
			
			//if (GL_SGIS_generate_mipmap_Flag)
				//glTexParameteri(tex.target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
			imgLoadImageAsTexture(img);
			
			glTexParameteri(tex.target, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(tex.target, GL_TEXTURE_WRAP_T, GL_REPEAT);
			//if (GL_SGIS_generate_mipmap_Flag)
				//glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			//else
				glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(tex.target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			
			imgDestroyImage(img);
			
			return textures.add(tex);
		}
	}
}



TextureID Renderer::addFunctionTexture2D(const unsigned int width, const unsigned int height, TEXTUREFUNC2D func, unsigned int nChannels){
	unsigned char *dest, *pixels = new unsigned char[width * height * nChannels];

	const float dx = 1.0f / float(width);
	const float dy = 1.0f / float(height);

	dest = pixels;
	for (unsigned int y = 0; y < height; y++){
		for (unsigned int x = 0; x < width; x++){
			unsigned int val = func(x * dx, y * dy);

			memcpy(dest, &val, nChannels);
			dest += nChannels;
		}
	}

	GLint format = (nChannels == 1)? GL_LUMINANCE : (nChannels == 3)? GL_RGB8 : GL_RGBA8;

	ShaderID shader = addMemoryTexture2D(width, height, pixels, format);
	delete pixels;
	return shader;	
}

TextureID Renderer::addFunctionTexture3D(const unsigned int width, const unsigned int height, const unsigned int depth, TEXTUREFUNC3D func, unsigned int nChannels){
	unsigned char *dest, *pixels = new unsigned char[width * height * depth * nChannels];

	const float dx = 1.0f / float(width);
	const float dy = 1.0f / float(height);
	const float dz = 1.0f / float(depth);

	dest = pixels;
	for (unsigned int z = 0; z < depth; z++){
		for (unsigned int y = 0; y < height; y++){
			for (unsigned int x = 0; x < width; x++){
				unsigned int val = func(x * dx, y * dy, z * dz);

				memcpy(dest, &val, nChannels);
				dest += nChannels;
			}
		}
	}

	GLint format = (nChannels == 1)? GL_LUMINANCE : (nChannels == 3)? GL_RGB8 : GL_RGBA8;

	ShaderID shader = addMemoryTexture3D(width, height, depth, pixels, format);
	delete pixels;
	return shader;	
}

TextureID Renderer::addMemoryTexture2D(const unsigned int width, const unsigned int height, void *pixels, GLint format, const bool clamp){
	Texture tex;
	glGenTextures(1, &tex.textureID);
	tex.target = GL_TEXTURE_2D;
	tex.isRenderTarget = false;
	glBindTexture(tex.target, tex.textureID);
	
	//if (GL_SGIS_generate_mipmap_Flag)
	//	glTexParameteri(tex.target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
	GLint type = (format == GL_LUMINANCE)? GL_LUMINANCE : (format == GL_RGB8)? GL_RGB : GL_RGBA;
	glTexImage2D(tex.target, 0, format, width, height, 0, type, GL_UNSIGNED_BYTE, pixels);

	glTexParameteri(tex.target, GL_TEXTURE_WRAP_S, clamp? GL_CLAMP_TO_EDGE_EXT : GL_REPEAT);
	glTexParameteri(tex.target, GL_TEXTURE_WRAP_T, clamp? GL_CLAMP_TO_EDGE_EXT : GL_REPEAT);
	//if (GL_SGIS_generate_mipmap_Flag)
		//glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	//else
		glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(tex.target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	return textures.add(tex);
}

TextureID Renderer::addMemoryTexture3D(const unsigned int width, const unsigned int height, const unsigned int depth, void *pixels, GLint format, const bool clamp){
	Texture tex;
	glGenTextures(1, &tex.textureID);
	tex.target = GL_TEXTURE_3D_EXT;
	tex.isRenderTarget = false;
	glBindTexture(tex.target, tex.textureID);
	
	//if (GL_SGIS_generate_mipmap_Flag)
	//	glTexParameteri(tex.target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

	GLint type = (format == GL_LUMINANCE)? GL_LUMINANCE : (format == GL_RGB8)? GL_RGB : GL_RGBA;
	glTexImage3DEXT(tex.target, 0, format, width, height, depth, 0, type, GL_UNSIGNED_BYTE, pixels);

	glTexParameteri(tex.target, GL_TEXTURE_WRAP_S,     clamp? GL_CLAMP_TO_EDGE_EXT : GL_REPEAT);
	glTexParameteri(tex.target, GL_TEXTURE_WRAP_T,     clamp? GL_CLAMP_TO_EDGE_EXT : GL_REPEAT);
	glTexParameteri(tex.target, GL_TEXTURE_WRAP_R_EXT, clamp? GL_CLAMP_TO_EDGE_EXT : GL_REPEAT);
	//if (GL_SGIS_generate_mipmap_Flag)
		//glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	//else
		glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(tex.target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	return textures.add(tex);
}

TextureID Renderer::addRenderTarget(int width, int height, int r, int g, int b, int a, int depth){
	Texture tex;
	glGenTextures(1, &tex.textureID);
	tex.target = GL_TEXTURE_2D;
	tex.isRenderTarget = true;
	glBindTexture(tex.target, tex.textureID);

	glTexParameteri(tex.target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
	glTexParameteri(tex.target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
	//if (GL_SGIS_generate_mipmap_Flag)
		//glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	//else
		glTexParameteri(tex.target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(tex.target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	int iAttributeIList[] = {
		WGL_SUPPORT_OPENGL_ARB,       GL_TRUE, 
		WGL_DRAW_TO_PBUFFER_ARB,      GL_TRUE,
		WGL_BIND_TO_TEXTURE_RGBA_ARB, GL_TRUE,

		WGL_PIXEL_TYPE_ARB,           WGL_TYPE_RGBA_ARB,
		WGL_ACCELERATION_ARB,         WGL_FULL_ACCELERATION_ARB,
		WGL_DOUBLE_BUFFER_ARB,        GL_FALSE,
		
		WGL_RED_BITS_ARB,             r,
		WGL_GREEN_BITS_ARB,           g,
		WGL_BLUE_BITS_ARB,            b,
		WGL_ALPHA_BITS_ARB,           a,
		WGL_DEPTH_BITS_ARB,           depth,

		0, 0
	};

    int pixelFormat;
	unsigned int nMatchingPixelFormats;
	wglChoosePixelFormatARB(hdc, iAttributeIList, NULL, 1, &pixelFormat, &nMatchingPixelFormats);
	
	int iPBufferAttributes[] = {
		WGL_TEXTURE_FORMAT_ARB, WGL_TEXTURE_RGBA_ARB,
		WGL_TEXTURE_TARGET_ARB, WGL_TEXTURE_2D_ARB,
		WGL_MIPMAP_TEXTURE_ARB, GL_FALSE,
		0, 0,
	};


	tex.hPBuffer = wglCreatePbufferARB(hdc, pixelFormat, width, height, iPBufferAttributes);
	tex.hPdc = wglGetPbufferDCARB(tex.hPBuffer);
	tex.hPrc = wglCreateContext(tex.hPdc);

	wglShareLists(hglrc, tex.hPrc);

	return textures.add(tex);
}



ShaderID Renderer::addShader(const char *vertexShaderFile, const char *fragmentShaderFile){
	const char *str;
	char *shaderText;
	int length;

	Shader shader;
	shader.programObj     = SHADER_NONE;
	shader.vertexShader   = SHADER_NONE;
	shader.fragmentShader = SHADER_NONE;

	GLhandle programObj = glCreateProgramObjectGL2();


	if (fragmentShaderFile != NULL){
		FILE *file = fopen(fragmentShaderFile, "rb");
		if (file == NULL){
			printf("Couldn't open %s\n", fragmentShaderFile);
		} else {
			fseek(file, 0, SEEK_END);
			length = ftell(file);
			fseek(file, 0, SEEK_SET);

			shaderText = new char[length];
			fread(shaderText, length, 1, file);
			fclose(file);
			
			GLuint fragmentShader = glCreateShaderObjectGL2(GL_FRAGMENT_SHADER_GL2);

			glLoadShaderGL2(fragmentShader, 1, (const char **) &shaderText, &length);
			if (glCompileShaderGL2(fragmentShader)){
				glAttachShaderObjectGL2(programObj, fragmentShader);
				shader.fragmentShader = fragmentShader;
			} else {
				printf("Fragment shader error:\n");
			}
			str = glGetInfoLogGL2(fragmentShader, 0);
			if (str && *str) printf("%s\n", str);
			//glDeleteObjectGL2(fragmentShader);

			delete shaderText;
		}
	}

	if (vertexShaderFile != NULL){
		FILE *file = fopen(vertexShaderFile, "rb");
		if (file == NULL){
			printf("Couldn't open %s\n", vertexShaderFile);
		} else {
			fseek(file, 0, SEEK_END);
			length = ftell(file);
			fseek(file, 0, SEEK_SET);

			shaderText = new char[length];
			fread(shaderText, length, 1, file);
			fclose(file);
			
			GLuint vertexShader = glCreateShaderObjectGL2(GL_VERTEX_SHADER_GL2);

			glLoadShaderGL2(vertexShader, 1, (const char **) &shaderText, &length);
			if (glCompileShaderGL2(vertexShader)){
				glAttachShaderObjectGL2(programObj, vertexShader);
				shader.vertexShader = vertexShader;
			} else {
				printf("Vertex shader error:\n");
			}
			str = glGetInfoLogGL2(vertexShader, 0);
			if (str && *str) printf("%s\n", str);
			//glDeleteObjectGL2(vertexShader);

			delete shaderText;
		}	
	}

	if (glLinkProgramGL2(programObj)){
		shader.programObj = programObj;
	} else {
		glDeleteObjectGL2(programObj);
	}
	str = glGetInfoLogGL2(programObj, 0);
	if (str && *str) printf("%s\n", str);

	if (shader.programObj != SHADER_NONE){
		return shaders.add(shader);
	} else {
		return SHADER_NONE;
	}
}

void Renderer::setRenderTarget(const TextureID texture){
	reset();
	apply();

	if (texture == TEXTURE_NONE){
		wglMakeCurrent(hdc, hglrc);
	} else {
	    wglMakeCurrent(textures[texture].hPdc, textures[texture].hPrc);
	}
}
