/*
	Emil Persson
	epersson@ati.com
	ATI Technologies Inc.

	A class for simplifying the handling of common stuff such as texture objects,
	shaders and blend modes.
*/

#ifndef _RENDERER_H_
#define _RENDERER_H_

#include <GL/glut.h>
#include "../include/ATiExtensions.h"
#include "Array.h"

#define TEXTURE_NONE (-1)
#define SHADER_NONE  (-1)

typedef unsigned int TextureID;
typedef unsigned int ShaderID;

#define MAX_TMU 4


struct Texture {
	GLuint textureID;
	GLuint target;

	// For render targets
	HPBUFFERARB hPBuffer;
	HDC hPdc;
	HGLRC hPrc;
	bool isRenderTarget;
};

struct Shader {
	GLhandle programObj;
	GLhandle vertexShader;
	GLhandle fragmentShader;
};

struct PixelFormat {
	unsigned int rBits;
	unsigned int gBits;
	unsigned int bBits;
	unsigned int aBits;
	unsigned int depthBits; 
};

typedef unsigned int (*TEXTUREFUNC2D)(const float s, const float t);
typedef unsigned int (*TEXTUREFUNC3D)(const float s, const float t, const float r);

unsigned int packf(const float r, const float g, const float b, const float a);


class Renderer {
protected:
	Array <Texture> textures;
	Array <Shader> shaders;

	TextureID selectedTextures[MAX_TMU];
	TextureID currentTextures[MAX_TMU];
	unsigned int nSelectedTextures, nCurrentTextures;

	ShaderID selectedShader;
	ShaderID currentShader;

	GLenum selectedSrcFactor, selectedDstFactor;
	GLenum currentSrcFactor,  currentDstFactor;
	bool selectedBlendEnabled;
	bool currentBlendEnabled;

	HDC hdc;
	HGLRC hglrc;

	// Resets the stored current state to defaults
	void defaults();
	// Resets the selected state to defaults
	void reset();
public:
	Renderer();
	~Renderer();

	void apply();

	TextureID addTexture(const char *fileName);

	TextureID addFunctionTexture2D(const unsigned int width, const unsigned int height, TEXTUREFUNC2D func, unsigned int nChannels = 4);
	TextureID addFunctionTexture3D(const unsigned int width, const unsigned int height, const unsigned int depth, TEXTUREFUNC3D func, unsigned int nChannels = 4);

	TextureID addMemoryTexture2D(const unsigned int width, const unsigned int height, void *pixels, GLint format = GL_RGBA8, const bool clamp = false);
	TextureID addMemoryTexture3D(const unsigned int width, const unsigned int height, const unsigned int depth, void *pixels, GLint format = GL_RGBA8, const bool clamp = false);

	TextureID addRenderTarget(int width, int height, int r, int g, int b, int a, int depth);

	ShaderID addShader(const char *vertexShaderFile, const char *fragmentShaderFile);

	GLhandle getProgram(const ShaderID shader = SHADER_NONE){
		if (shader == SHADER_NONE){
			return (currentShader != SHADER_NONE)? shaders[currentShader].programObj : 0;
		} else return shaders[shader].programObj;
	}

	void setRenderTarget(const TextureID texture);

	void setBlending(const GLenum srcFactor, const GLenum dstFactor){
		selectedSrcFactor = srcFactor;
		selectedDstFactor = dstFactor;
		selectedBlendEnabled = true;
	}

	void setTextureAt(const unsigned int index, const TextureID texture){
		selectedTextures[index] = texture;
		if (index >= nSelectedTextures) nSelectedTextures = index + 1;
	}

	void setTextures(const TextureID t0){
		selectedTextures[0] = t0;
		nSelectedTextures = 1;
	}
	void setTextures(const TextureID t0, const TextureID t1){
		selectedTextures[0] = t0;
		selectedTextures[1] = t1;
		nCurrentTextures = 2;
	}
	void setShader(const ShaderID shader){
		selectedShader = shader;
	}

};



#endif // _RENDERER_H_
